function [fval info dif_vec GQ AQ mu] = blgaussexp(A,W,opts)
%BLGAUSSEXP Compute W'*expm(A)*W via block Gauss/anti-Gauss quadrature.
%   fval = BLGAUSSEXP(A,W) computes W'*expm(A)*W, for a symmetric matrix A and
%   an nxk block W, using Gauss and anti-Gauss quadrature.
%
%   fval = BLGAUSSEXP(A,W,opts) adds more information.
%	opts: additional options, see code.
%
%   [fval info] = BLGAUSSEXP(A,...) gets more information.
%	fval: approximation of W'*expm(A)*W.
%	info: info(1) is always 0. info(2) contains the number of Lanczos steps
%	      info(3) contains the number of matrix-block products.

%   Caterina Fenu & Giuseppe Rodriguez, University of Cagliari, Italy
%   David Martin & Lothar Reichel, Kent State University, U.S.A.
%   Email: kate.fenu@gmail.com, rodriguez@unica.it, 
%          dmarti49@kent.edu, reichel@math.kent.edu
%
%   Last revised August 8, 2013

if nargin < 3 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 2, error('Too few input arguments.'); end

[n bsize] = size(W);
if isnumeric(A)
	if ~isequal(A,A'), error('A must be symmetric.'), end
	mmprod = @(X) A*X;
elseif isa(A,'char') || isa(A,'function_handle') || isa(A,'inline')
	mmprod = A;
else
	error('A must be either a numeric matrix or a function.')
end

nam = fieldnames(opts);
% tolerance for quadrature convergence
if ~any(strcmpi('gausstolq',nam)), opts.gausstolq = 1e-3; end
% tolerance for check if beta is 0
if ~any(strcmpi('gausstau',nam)), opts.gausstau = 1e-10; end
% max number of iterations
if ~any(strcmpi('gaussmaxn',nam)), opts.gaussmaxn = min(100,n); end
% value of mu for scaling expm, defaults to automatic
if ~any(strcmpi('gaussmu',nam)), opts.gaussmu = -1; end
% 1 for reorth (see code)
if ~any(strcmpi('gaussreorth',nam)), opts.gaussreorth = 1; end
% 1 for mgs
if ~any(strcmpi('gaussmgs',nam)), opts.gaussmgs = 1; end

mu = opts.gaussmu;
automu = (mu<0);	% a negative value activates automatic spectral shift

%expm_fun = @(T,mu,U,D) expm(T-mu*eye(size(T,1)));
expm_fun = @(T,mu,U,D) expm_eig(T,mu,U,D);

% check if the columns of block W are orthonormal
rescale = 0;
if norm(W'*W-eye(bsize),inf) > sqrt(eps)
	rescale = 1;
	wsparse = 0;
	%warning('blgaussexp:orth','The block W is not orthonormal. It will be replaced through QR factorization');
	if wsparse
		[W RR] = qr(W,0);
	else
		[Ut dt] = eig(W'*W);
		sqdt = sqrt(diag(dt));
		W = W*Ut*diag(1./sqdt);
		RR = diag(sqdt)*Ut';
	end
end

if opts.gaussreorth, Q = zeros(n,opts.gaussmaxn*bsize);end

% starting vectors
G = zeros(opts.gaussmaxn*bsize,opts.gaussmaxn*bsize);
%E = zeros(opts.gaussmaxn*bsize,bsize);		% to check computation
gamma = zeros(bsize,bsize);
X0 = zeros(n,bsize);
X1 = W;

if opts.gaussreorth
	Q(:,1:bsize) = X1;
end

flag = 1;
i = 0;
info = 0;
sqrt_two = sqrt(2);

i_b = 0;
i = i+1;	% number of iterations = number of matrix-block products
temp = mmprod(X1);
omega = X1'*temp;
G(1:bsize,1:bsize) = omega;
%E(1:bsize,1:bsize) = eye(bsize);		% to check computation
dif_vec = zeros(opts.gaussmaxn,1);

while flag && (i<opts.gaussmaxn)

	i = i + 1;

	R = temp - X1*omega - X0*gamma';

	if opts.gaussreorth
		for j=1:i
			ind_a = (j-1)*bsize+1;
			ind_b = j*bsize;
			R=R-Q(:,ind_a:ind_b)*(Q(:,ind_a:ind_b)'*R);
			%keyboard
			%errore = A*Q - Q*G - X1*gamma*E'% to check computation
		end
	end

	X0 = X1;
	[X1,gamma]=qr(R,0);
	if min(abs(diag(gamma))) < opts.gausstau
		warning(sprintf('gamma block singular at iteration %d (%.2e)\n', i, min(abs(diag(gamma)))))
	end
	temp = mmprod(X1);
	omega = X1'*temp;

	i_a = i_b + 1;		% (i-1)*bsize+1
	i_b = i_b + bsize;	% i*bsize

	if opts.gaussreorth
		Q(:,i_a+bsize:i_b+bsize) = X1;
	end

	T = G(1:i_b,1:i_b);
	if automu
		[U1 D1] = eig(T);
		mu = max(diag(D1));	% Gauss spectral shift
	else
		U1 = [];  D1 = [];
	end

	G(i_a:i_b,i_b+1:i_b+bsize) = sqrt_two*gamma';
	G(i_b+1:i_b+bsize,i_a:i_b) = sqrt_two*gamma;
	G(i_b+1:i_b+bsize,i_b+1:i_b+bsize) = omega;

	% spectral shift is max of Gauss/anti-Gauss spectral shifts
	if automu
		[U2 D2] = eig(G(1:i_b+bsize,1:i_b+bsize));
		mu = max( mu, max(diag(D2)));
	else
		U2 = [];  D2 = [];
	end

	FG = expm_fun(G(1:i_b,1:i_b),mu,U1,D1);
	GQ = FG(1:bsize,1:bsize);

	FH = expm_fun(G(1:i_b+bsize,1:i_b+bsize),mu,U2,D2);
	AQ = FH(1:bsize,1:bsize);

	G(i_a:i_b,i_b+1:i_b+bsize) = gamma';
	G(i_b+1:i_b+bsize,i_a:i_b) = gamma;

	% Half the difference between the Gauss and anti-Gauss rules.
	dif = max(max((abs(GQ-AQ))))/2;
	den = max(max((abs(GQ))));
	dif_vec(i) = dif/den;
	flag = dif > opts.gausstolq*den;

end

dif_vec = dif_vec(1:i);
fval = (GQ + AQ)/2;
if rescale
	fval = RR'*fval*RR;
end

e = exp(mu);
if isfinite(e)		% recover correct values from spectral shift
	fval = e*fval;
else
	warning('blgaussexp:scaledCentrality',...
		'Subgraph centrality is scaled by exp(%.2e).',mu)
end

info = [info i bsize*i];


function F = expm_eig(T,mu,V,D)

if nargin < 2, mu = 0; end
if nargin < 4 || isempty(V), [V D] = eig(T); end

F = V*diag(exp(diag(D)-mu))*V';

