function [fval info dif_vec GQ AQ mu] = blgaussexpns(A,W,V,opts)
%BLGAUSSEXPNS Compute W'*expm(A)*V via block Gauss/anti-Gauss quadrature.
%   fval = BLGAUSSEXPNS(A,W,V) computes W'*expm(A)*V, for a nonsymmetric matrix
%   A and two nxk blocks W,V using Gauss and anti-Gauss quadrature.
%
%   fval = BLGAUSSEXPNS(A,W,V,opts) adds more information.
%	opts: additional options, see code.
%
%   [fval info] = BLGAUSSEXPNS(A,...) gets more information.
%	fval: approximation of W'*expm(A)*W.
%	info: info(1) is always 0. info(2) contains the number of Lanczos steps
%	      info(3) contains the number of matrix-block products,
%	      info(4) contains the number of matrix-vectors products.

%   Caterina Fenu & Giuseppe Rodriguez, University of Cagliari, Italy
%   David Martin & Lothar Reichel, Kent State University, U.S.A.
%   Email: kate.fenu@gmail.com, rodriguez@unica.it, 
%          dmarti49@kent.edu, reichel@math.kent.edu
%
%   Last revised August 4, 2013

if nargin < 4 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 3, error('Too few input arguments.'); end

[n bsize] = size(W);
[n2 bsize2] = size(V);
if n~=n2 || bsize ~=bsize2
	error('The blocks are of different size');
end

if isnumeric(A)
	mmprod = @(X) A*X;
	mmtprod = @(X) A'*X;
elseif iscell(A)
	mmprod = A{1};
	mmtprod = A{2};
else
	error('A must be either a numeric matrix or a cell array.')
end

nam = fieldnames(opts);
% tolerance for quadrature convergence
if ~any(strcmpi('gausstolq',nam)), opts.gausstolq = 1e-3; end
% tolerance for check if beta is 0
if ~any(strcmpi('gausstau',nam)), opts.gausstau = 1e-10; end
% max number of iterations
if ~any(strcmpi('gaussmaxn',nam)), opts.gaussmaxn = min(100,n); end
% value of mu for scaling expm
if ~any(strcmpi('gaussmu',nam)), opts.gaussmu = -1; end
% 1 for reorth (see code)
if ~any(strcmpi('gaussreorth',nam)), opts.gaussreorth = 1; end
% 1 for mgs
if ~any(strcmpi('gaussmgs',nam)), opts.gaussmgs = 1; end

mu = opts.gaussmu;
automu = (mu<0);	% a negative value activates automatic spectral shift

expm_fun = @(T,mu,U,D) expm(T-mu*eye(size(T,1)));
%expm_fun = @(T,mu,U,D) expm_eig(T,mu,U,D);

% check if the columns of block W are orthonormal
rescale = 0;
if norm(W'*V-eye(bsize),inf) > sqrt(eps)
	rescale = 1;
	%warning('blgaussexpns:orth','The block W''V is not orthonormal. It will be replaced through svd');
	[Ut sigmat Zt] = svd(W'*V);
	sqrSigt = diag(sqrt(diag(sigmat)));
	isqrSigt = diag(sqrt(1./diag(sigmat)));
	W = W*Ut*isqrSigt;
	V = V*Zt*isqrSigt;
end

if opts.gaussreorth
	Q1 = zeros(n,opts.gaussmaxn*bsize);
	Q2 = zeros(n,opts.gaussmaxn*bsize);
end

% starting vectors
G = zeros(opts.gaussmaxn*bsize,opts.gaussmaxn*bsize);
%E = zeros(opts.gaussmaxn*bsize,bsize);		% to check computation
gamma = zeros(bsize,bsize);
delta = zeros(bsize,bsize);
V0 = zeros(n,bsize);
W0 = zeros(n,bsize);
V1 = V;
W1 = W;

if opts.gaussreorth
	Q1(:,1:bsize) = V1;
	Q2(:,1:bsize) = W1;
end

flag = 1;
i = 0;
info = 0;
sqrt_two = sqrt(2);
nprods = 0;		% number of matrix-block products

i_b = 0;
temp = mmprod(V1)-V0*delta';  nprods = nprods+1;
i = i + 1;
omega = W1'*temp;
G(1:bsize,1:bsize) = omega;
%E(1:bsize,1:bsize) = eye(bsize);		% to check computation
dif_vec = zeros(opts.gaussmaxn,1);

while flag && (i<opts.gaussmaxn)

	i = i + 1;

	R = temp - V1*omega;
	S = mmtprod(W1) - W1*omega' - W0*gamma';
	nprods = nprods+1;

	V0 = V1;
	W0 = W1;

	if (rank(R) < bsize) || (rank(S) < bsize)
		% we need to add code here
		warning('blgaussexpns:rankdefR','either R or S is rank deficient at step %d',i);
	elseif (rank(S'*R) < bsize)
		error('blgaussexpns:seriousbreakdown','serious breakdown at step %d',i); 
	end

	if opts.gaussreorth
		for j=1:i
			ind_a = (j-1)*bsize+1;
			ind_b = j*bsize;
			R = R-Q1(:,ind_a:ind_b)*(Q2(:,ind_a:ind_b)'*R);
			S = S-Q2(:,ind_a:ind_b)*(Q1(:,ind_a:ind_b)'*S);
			%keyboard
			%errore = A*Q - Q*G - X1*gamma*E'% to check computation
		end
	end

	[QR,RR] = qr(R,0);
	[QS,RS] = qr(S,0);
	[U sigma Z] = svd(QS'*QR);
	sqrSig = diag(sqrt(diag(sigma)));
	isqrSig = diag(sqrt(1./diag(sigma)));
	V1 = QR*Z*isqrSig;
	W1 = QS*U*isqrSig;

	gamma = sqrSig*Z'*RR;
	delta = sqrSig*U'*RS;

	temp = mmprod(V1) - V0*delta';
	nprods = nprods+1;
	omega = W1'*temp;

	i_a = i_b + 1;		% (i-1)*bsize+1
	i_b = i_b + bsize;	% i*bsize

	if opts.gaussreorth
		Q1(:,i_a+bsize:i_b+bsize) = V1;
		Q2(:,i_a+bsize:i_b+bsize) = W1;
	end

	T = G(1:i_b,1:i_b);
	if automu		% Gauss spectral shift
		[U1 D1] = eig(T);
		mu = max(real(diag(D1)));
	else
		U1 = [];  D1 = [];
	end

	G(i_a:i_b,i_b+1:i_b+bsize) = sqrt_two*delta';
	G(i_b+1:i_b+bsize,i_a:i_b) = sqrt_two*gamma;
	G(i_b+1:i_b+bsize,i_b+1:i_b+bsize) = omega;

	% spectral shift is max of Gauss/anti-Gauss spectral shifts
	if automu
		[U2 D2] = eig(G(1:i_b+bsize,1:i_b+bsize));
		mu = max( mu, max(real(diag(D2))));
	else
		U2 = [];  D2 = [];
	end

	FG = expm_fun(T,mu,U1,D1);
	GQ = FG(1:bsize,1:bsize);

	FH = expm_fun(G(1:i_b+bsize,1:i_b+bsize),mu,U2,D2);
	AQ = FH(1:bsize,1:bsize);

	G(i_a:i_b,i_b+1:i_b+bsize) = delta';
	G(i_b+1:i_b+bsize,i_a:i_b) = gamma;

	% Half the difference between the Gauss and anti-Gauss rules.
	dif = max(max(abs(GQ-AQ)))/2;
	den = max(max(abs(GQ)));
	dif_vec(i) = dif/den;
	flag = dif > opts.gausstolq*den;

end

dif_vec = dif_vec(1:i);
fval = (GQ + AQ)/2;
if rescale
	%warning('blgaussexpns:rescale','The value of W''*expm(A)*V is not rescaled properly');
	fval = Ut*sqrSigt*fval*sqrSigt*Zt';
	GQ = Ut*sqrSigt*GQ*sqrSigt*Zt';
	AQ = Ut*sqrSigt*AQ*sqrSigt*Zt';
end

e = exp(mu);
if isfinite(e)		% recover correct values from spectral shift
	fval = e*fval;
else
	warning('blgaussexpns:scaledCentrality',...
		'Subgraph centrality is scaled by exp(%.2e).',mu)
end

info = [info i nprods bsize*nprods];


function F = expm_eig(T,mu,V,D)

if nargin < 2, mu = 0; end
if nargin < 4 || isempty(V), [V D] = eig(T); end

if isequal(T,T')
	F = V*diag(exp(diag(D)-mu))*V';
else
	F = V*diag(exp(diag(D)-mu))/V;
end

